/*
  Copyright 2007 Darren Watts

  This file is part of JUICE.

  JUICE is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  JUICE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with JUICE; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

package src.gui;
import src.backend.*;
import javax.swing.table.AbstractTableModel;

public class LinsTableModel extends AbstractTableModel{
    public static final int LINS_INDEX = 0;
    public static final int LINS_SOLID = 1;
    public static final int LINS_TRANS = 2;
    public static final int LINS_LANDSCAPED = 3;
    public static final int LINS_LENGTH = 4;
    public static final int LINS_HIGHFLOOR = 5;
    public static final int LINS_LOWCEILING = 6;
    public static final int LINS_CPOWNER = 7;
    public static final int LINS_CCPOWNER = 8;

    public static final int FLAG_SOLID = 0x4000;
    public static final int FLAG_TRANSPARENT = 0x2000;
    public static final int FLAG_LANDSCAPED = 0x1000;

    public static final int FLAG_NOT_SOLID = 0x3FFF;
    public static final int FLAG_NOT_TRANSPARENT = 0x5FFF;
    public static final int FLAG_NOT_LANDSCAPED = 0x6FFF;

    private String[] colNames;
    private LinsEntry[] data;
    private int[] indexNums;
    private Model model;

    public LinsTableModel(String[] colNames, Model model, int levelIndex) {
        this.colNames = colNames;
        this.model = model;
        data = new LinsEntry[this.model.getLinsNumber(levelIndex)];
        indexNums = new int[this.model.getLinsNumber(levelIndex)];

        for(int i = 0; i < this.model.getLinsNumber(levelIndex); i++){
            indexNums[i] = i;
            data[i] = this.model.getLin(levelIndex, i);
        }
    }

    public String getColumnName(int col){
        return colNames[col];
    }

    public boolean isCellEditable(int row, int column) {
        if(column == LINS_INDEX) return false;
        else if (column == LINS_LENGTH && !model.isOptimized()) return false;
        else if(column == LINS_HIGHFLOOR && !model.isOptimized()) return false;
        else if(column == LINS_LOWCEILING && !model.isOptimized()) return false;
        else if(column == LINS_CPOWNER) return false;
        else if(column == LINS_CCPOWNER) return false;
        else return true;
    }

    public Object getValueAt(int row, int column) {
        Integer num;
        switch (column) {
        case LINS_INDEX:
            return new Integer(indexNums[row]);
        case LINS_SOLID:
            return new Boolean((data[row].getFlags() & FLAG_SOLID) > 0);
        case LINS_TRANS:
            return new Boolean((data[row].getFlags() & FLAG_TRANSPARENT) > 0);
        case LINS_LANDSCAPED:
            return new Boolean((data[row].getFlags() & FLAG_LANDSCAPED) > 0);
        case LINS_LENGTH:
            return new Integer(data[row].getLength());
        case LINS_HIGHFLOOR:
            return new Integer(data[row].getHighestAdjacentFloor());
        case LINS_LOWCEILING:
            return new Integer(data[row].getLowestAdjacentCeiling());
        case LINS_CPOWNER:
            return new Integer(data[row].getClockwisePolyOwner());
        case LINS_CCPOWNER:
            return new Integer(data[row].getCounterClockwisePolyOwner());
        default:
            return new Object();
        }
   }

   public void setValueAt(Object value, int row, int column) {
       Integer tempValue;
       boolean boolValue = false;
       try {
           boolValue = ((Boolean) value).booleanValue();
       } catch (Exception ex) {}

       switch (column) {
       case LINS_SOLID:
           if (boolValue) data[row].setFlags(data[row].getFlags() | FLAG_SOLID);
           else data[row].setFlags(data[row].getFlags() & FLAG_NOT_SOLID);
           break;
       case LINS_TRANS:
           if (boolValue) data[row].setFlags(data[row].getFlags() | FLAG_TRANSPARENT);
           else data[row].setFlags(data[row].getFlags() & FLAG_NOT_TRANSPARENT);
           break;
       case LINS_LANDSCAPED:
           if (boolValue) data[row].setFlags(data[row].getFlags() | FLAG_LANDSCAPED);
           else data[row].setFlags(data[row].getFlags() & FLAG_NOT_LANDSCAPED);
           break;
       case LINS_LENGTH:
           tempValue = Integer.valueOf(((String) value));
           data[row].setLength(tempValue.shortValue());
           break;
       case LINS_HIGHFLOOR:
           tempValue = Integer.valueOf(((String) value));
           data[row].setHighestAdjacentFloor(tempValue.shortValue());
           break;
       case LINS_LOWCEILING:
           tempValue = Integer.valueOf(((String) value));
           data[row].setLowestAdjacentCeiling(tempValue.shortValue());
           break;
       default:
           System.out.println("Invalid LinsTable index");
       }
       fireTableCellUpdated(row, column);
   }

   public int getRowCount() {
       return data.length;
   }

   public int getColumnCount(){
       return colNames.length;
   }

   public void sort(int header, boolean reverse){
       try{
           switch (header) {
           case LINS_INDEX:
               sortBy(reverse, "index");
               break;
           case LINS_SOLID:
               sortBy(reverse, "solid");
               break;
           case LINS_TRANS:
               sortBy(reverse, "trans");
               break;
           case LINS_LANDSCAPED:
               sortBy(reverse, "landscaped");
               break;
           case LINS_LENGTH:
               sortBy(reverse, "length");
               break;
           case LINS_HIGHFLOOR:
               sortBy(reverse, "hf");
               break;
           case LINS_LOWCEILING:
               sortBy(reverse, "lc");
               break;
           case LINS_CPOWNER:
               sortBy(reverse, "cpowner");
               break;
           case LINS_CCPOWNER:
               sortBy(reverse, "ccpowner");
               break;
           }
           this.fireTableDataChanged();
       } catch(Exception e){}
   }

   private void sortBy(boolean reverse, String type){
       mergeSort(0, data.length, reverse, type);
   }

   private void mergeSort(int first, int n, boolean reverse, String type){
       int n1; // Size of the first half of the array
       int n2; // Size of the second half of the array

       if (n > 1) {
           // Compute sizes of the two halves
           n1 = n / 2;
           n2 = n - n1;

           mergeSort(first, n1, reverse, type); // Sort data[first] through data[first+n1-1]
           mergeSort(first + n1, n2, reverse, type); // Sort data[first+n1] to the end

           // Merge the two sorted halves.
           merge(first, n1, n2, reverse, type);
       }
   }

   private void merge(int first, int n1, int n2, boolean reverse, String type) {
       int[] temp = new int[n1 + n2]; // Allocate the temporary array
       LinsEntry[] tempEntry = new LinsEntry[n1 + n2];
       int copied = 0; // Number of elements copied from data to temp
       int copied1 = 0; // Number copied from the first half of data
       int copied2 = 0; // Number copied from the second half of data
       int i; // Array index to copy from temp back into data

       // Merge elements, copying from two halves of data to the temporary array.
       while ((copied1 < n1) &&
              (copied2 < n2)) {
           if (compare(reverse, type, first, copied1, copied2, n1)) {
               temp[copied++] = indexNums[first + (copied1++)];
               tempEntry[copied - 1] = data[first + copied1 - 1];
           } else {
               temp[copied++] = indexNums[first + n1 + (copied2++)];
               tempEntry[copied - 1] = data[first + n1 + copied2 - 1];
           }
       }

       // Copy any remaining entries in the left and right subarrays.
       while (copied1 < n1) {
           temp[copied++] = indexNums[first + (copied1++)];
           tempEntry[copied - 1] = data[first + copied1 - 1];
       } while (copied2 < n2) {
           temp[copied++] = indexNums[first + n1 + (copied2++)];
           tempEntry[copied - 1] = data[first + n1 + copied2 - 1];
       }
       // Copy from temp back to the data array.
       for (i = 0; i < n1 + n2; i++) {
           indexNums[first + i] = temp[i];
           data[first + i] = tempEntry[i];
       }
   }


   private boolean compare(boolean reverse, String type, int first,
                           int copied1, int copied2, int n1) {
       if (!reverse && type.equals("index")) {
           return (indexNums[first + copied1] < indexNums[first + n1 + copied2]);
       }
       else if (reverse && type.equals("index")) {
           return (indexNums[first + copied1] > indexNums[first + n1 + copied2]);
       }
       else if (!reverse && type.equals("solid")) {
           return (data[first + copied1].getFlags() & FLAG_SOLID) > 0;
       }
       else if (reverse && type.equals("solid")){
           return (data[first + copied1].getFlags() & FLAG_SOLID) == 0;
       }
       else if (!reverse && type.equals("transparent")) {
           return (data[first + copied1].getFlags() & FLAG_TRANSPARENT) > 0;
       }
       else if (reverse && type.equals("transparent")){
           return (data[first + copied1].getFlags() & FLAG_TRANSPARENT) == 0;
       }
       else if (!reverse && type.equals("landscaped")) {
           return (data[first + copied1].getFlags() & FLAG_LANDSCAPED) > 0;
       }
       else if (reverse && type.equals("landscaped")){
           return (data[first + copied1].getFlags() & FLAG_LANDSCAPED) == 0;
       }
       else if (!reverse && type.equals("length")) {
           return data[first + copied1].getLength() <
                   data[first + n1 + copied2].getLength();
       }
       else if (reverse && type.equals("length")){
           return data[first + copied1].getLength() >
                   data[first + n1 + copied2].getLength();
       }
       else if (!reverse && type.equals("hf")) {
           return data[first + copied1].getHighestAdjacentFloor() <
                   data[first + n1 + copied2].getHighestAdjacentFloor();
       }
       else if (reverse && type.equals("hf")){
           return data[first + copied1].getHighestAdjacentFloor() >
                   data[first + n1 + copied2].getHighestAdjacentFloor();
       }
       else if (!reverse && type.equals("lc")) {
           return data[first + copied1].getLowestAdjacentCeiling() <
                   data[first + n1 + copied2].getLowestAdjacentCeiling();
       }
       else if (reverse && type.equals("lc")){
           return data[first + copied1].getLowestAdjacentCeiling() >
                   data[first + n1 + copied2].getLowestAdjacentCeiling();
       }
       else if (!reverse && type.equals("cpowner")) {
           return data[first + copied1].getClockwisePolyOwner() <
                   data[first + n1 + copied2].getClockwisePolyOwner();
       }
       else if (reverse && type.equals("cpowner")){
           return data[first + copied1].getClockwisePolyOwner() >
                   data[first + n1 + copied2].getClockwisePolyOwner();
       }
       else if (!reverse && type.equals("ccpowner")) {
           return data[first + copied1].getCounterClockwisePolyOwner() <
                   data[first + n1 + copied2].getCounterClockwisePolyOwner();
       }
       else if (reverse && type.equals("ccpowner")){
           return data[first + copied1].getCounterClockwisePolyOwner() >
                   data[first + n1 + copied2].getCounterClockwisePolyOwner();
       }
       else return false;
   }
}
